@@ -37,6 +37,8 @@
 
 #define P11_DEBUG_FLAG P11_DEBUG_RPC
 #include "debug.h"
+#include "filter.h"
+#include "iter.h"
 #include "pkcs11.h"
 #include "library.h"
 #include "private.h"
@@ -44,6 +46,7 @@
 #include "remote.h"
 #include "rpc.h"
 #include "rpc-message.h"
+#include "uri.h"
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -280,11 +283,7 @@ proto_read_attribute_array (p11_rpc_message *msg,
                             CK_ULONG *n_result)
 {
 	CK_ATTRIBUTE_PTR attrs;
-	const unsigned char *data;
-	unsigned char valid;
 	uint32_t n_attrs, i;
-	uint32_t value;
-	size_t n_data;
 
 	assert (msg != NULL);
 	assert (result != NULL);
@@ -305,34 +304,32 @@ proto_read_attribute_array (p11_rpc_message *msg,
 
 	/* Now go through and fill in each one */
 	for (i = 0; i < n_attrs; ++i) {
+		size_t offset = msg->parsed;
+		CK_ATTRIBUTE temp;
 
-		/* The attribute type */
-		if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &value))
+		/* Check the length needed to store the value */
+		memset (&temp, 0, sizeof (temp));
+		if (!p11_rpc_buffer_get_attribute (msg->input, &offset, &temp)) {
+			msg->parsed = offset;
 			return PARSE_ERROR;
+		}
 
-		attrs[i].type = value;
+		attrs[i].type = temp.type;
 
 		/* Whether this one is valid or not */
-		if (!p11_rpc_buffer_get_byte (msg->input, &msg->parsed, &valid))
-			return PARSE_ERROR;
-
-		if (valid) {
-			if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &value))
-				return PARSE_ERROR;
-			if (!p11_rpc_buffer_get_byte_array (msg->input, &msg->parsed, &data, &n_data))
-				return PARSE_ERROR;
-
-			if (data != NULL && n_data != value) {
-				p11_message ("attribute length and data do not match");
+		if (temp.ulValueLen != ((CK_ULONG)-1)) {
+			size_t offset2 = msg->parsed;
+			attrs[i].pValue = p11_rpc_message_alloc_extra (msg, temp.ulValueLen);
+			if (!p11_rpc_buffer_get_attribute (msg->input, &offset2, &attrs[i])) {
+				msg->parsed = offset2;
 				return PARSE_ERROR;
 			}
-
-			attrs[i].pValue = (CK_VOID_PTR)data;
-			attrs[i].ulValueLen = value;
 		} else {
 			attrs[i].pValue = NULL;
 			attrs[i].ulValueLen = -1;
 		}
+
+		msg->parsed = offset;
 	}
 
 	*result = attrs;
@@ -405,9 +402,8 @@ static CK_RV
 proto_read_mechanism (p11_rpc_message *msg,
                       CK_MECHANISM_PTR mech)
 {
-	const unsigned char *data;
-	uint32_t value;
-	size_t n_data;
+	size_t offset;
+	CK_MECHANISM temp;
 
 	assert (msg != NULL);
 	assert (mech != NULL);
@@ -416,17 +412,31 @@ proto_read_mechanism (p11_rpc_message *msg,
 	/* Make sure this is in the right order */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "M"));
 
-	/* The mechanism type */
-	if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &value))
+	/* Check the length needed to store the parameter */
+	memset (&temp, 0, sizeof (temp));
+	offset = msg->parsed;
+	if (!p11_rpc_buffer_get_mechanism (msg->input, &offset, &temp)) {
+		msg->parsed = offset;
+		return PARSE_ERROR;
+	}
+
+	mech->mechanism = temp.mechanism;
+
+	/* The mechanism doesn't require parameter */
+	if (temp.ulParameterLen == 0) {
+		mech->pParameter = NULL;
+		mech->ulParameterLen = 0;
+		msg->parsed = offset;
+		return CKR_OK;
+	}
+
+	/* Actually retrieve the parameter */
+	mech->pParameter = p11_rpc_message_alloc_extra (msg, temp.ulParameterLen);
+	if (!p11_rpc_buffer_get_mechanism (msg->input, &msg->parsed, mech))
 		return PARSE_ERROR;
 
-	/* The mechanism data */
-	if (!p11_rpc_buffer_get_byte_array (msg->input, &msg->parsed, &data, &n_data))
-		return PARSE_ERROR;
+	assert (msg->parsed == offset);
 
-	mech->mechanism = value;
-	mech->pParameter = (CK_VOID_PTR)data;
-	mech->ulParameterLen = n_data;
 	return CKR_OK;
 }
 
@@ -1916,6 +1926,16 @@ p11_rpc_server_handle (CK_X_FUNCTION_LIST *self,
 	return true;
 }
 
+/**
+ * p11_kit_remote_serve_module:
+ * @module: a pointer to a loaded module
+ * @in_fd: input fd
+ * @out_fd: output fd
+ *
+ * Run a module on a given pair of input/output FDs.
+ *
+ * Returns: 0 if success, non-zero otherwise.
+ */
 int
 p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
                              int in_fd,
@@ -1952,7 +1972,7 @@ p11_kit_remote_serve_module (CK_FUNCTION_LIST *module,
 	}
 
 	version = 0;
-	switch (write (out_fd, &version, out_fd)) {
+	switch (write (out_fd, &version, 1)) {
 	case 1:
 		break;
 	default:
@@ -2015,3 +2035,169 @@ out:
 
 	return ret;
 }
+
+/**
+ * p11_kit_remote_serve_token:
+ * @module: a pointer to a loaded module
+ * @token: a token info
+ * @in_fd: input fd
+ * @out_fd: output fd
+ *
+ * Run a token wrapped in a module on a given pair of input/output FDs.
+ *
+ * Returns: 0 if success, non-zero otherwise.
+ *
+ * Deprecated: use p11_kit_remote_serve_tokens()
+ */
+int
+p11_kit_remote_serve_token (CK_FUNCTION_LIST *module,
+			    CK_TOKEN_INFO *token,
+			    int in_fd,
+			    int out_fd)
+{
+	p11_virtual virt;
+	p11_virtual *filter = NULL;
+	CK_FUNCTION_LIST *filtered = NULL;
+	int ret = 1;
+
+	return_val_if_fail (module != NULL, 1);
+	return_val_if_fail (token != NULL, 1);
+
+	p11_virtual_init (&virt, &p11_virtual_base, module, NULL);
+	filter = p11_filter_subclass (&virt, NULL);
+	if (filter == NULL)
+		goto out;
+
+	filtered = p11_virtual_wrap (filter, (p11_destroyer)p11_virtual_uninit);
+	if (filtered == NULL)
+		goto out;
+
+	p11_filter_allow_token (filter, token);
+
+	ret = p11_kit_remote_serve_module (filtered, in_fd, out_fd);
+
+ out:
+	if (filtered != NULL)
+		p11_virtual_unwrap (filtered);
+	if (filter != NULL)
+		p11_filter_release (filter);
+
+	return ret;
+}
+
+/* Find a module providing a token */
+static CK_FUNCTION_LIST *
+find_module (const char *token)
+{
+	CK_FUNCTION_LIST **modules = NULL, *module = NULL;
+	P11KitUri *uri = NULL;
+	P11KitIter *iter = NULL;
+	CK_RV rv;
+
+	modules = p11_kit_modules_load_and_initialize (0);
+	if (modules == NULL)
+		goto out;
+
+	uri = p11_kit_uri_new ();
+	if (uri == NULL)
+		goto out;
+
+	if (p11_kit_uri_parse (token, P11_KIT_URI_FOR_TOKEN, uri) !=
+	    P11_KIT_URI_OK)
+		goto out;
+
+	iter = p11_kit_iter_new (uri,
+				 P11_KIT_ITER_WITH_TOKENS |
+				 P11_KIT_ITER_WITHOUT_OBJECTS);
+	p11_kit_uri_free (uri);
+	uri = NULL;
+	if (iter == NULL)
+		goto out;
+
+	p11_kit_iter_begin (iter, modules);
+	rv = p11_kit_iter_next (iter);
+	if (rv != CKR_OK)
+		goto out;
+
+	module = p11_kit_iter_get_module (iter);
+
+ out:
+	p11_kit_iter_free (iter);
+	p11_kit_modules_release (modules);
+
+	return module;
+}
+
+/**
+ * p11_kit_remote_serve_tokens:
+ * @tokens: a list of token URIs
+ * @n_tokens: the length of @tokens
+ * @module: (nullable): a PKCS\#11 module that provides the tokens
+ * @in_fd: input fd
+ * @out_fd: output fd
+ *
+ * Expose tokens on a given pair of input/output FDs.  All the tokens
+ * must be provided by the same module.
+ *
+ * Returns: 0 if success, non-zero otherwise.
+ */
+int
+p11_kit_remote_serve_tokens (const char **tokens,
+			     size_t n_tokens,
+			     CK_FUNCTION_LIST *module,
+			     int in_fd,
+			     int out_fd)
+{
+	p11_virtual virt;
+	p11_virtual *filter = NULL;
+	CK_FUNCTION_LIST *filtered = NULL;
+	int ret = 1;
+	size_t i;
+
+	return_val_if_fail (tokens != NULL, 2);
+	return_val_if_fail (n_tokens > 0, 2);
+	return_val_if_fail (in_fd >= 0, 2);
+	return_val_if_fail (out_fd >= 0, 2);
+
+	if (!module)
+		module = find_module (tokens[0]);
+
+	/* Create a virtual module that provides only the specified tokens */
+	p11_virtual_init (&virt, &p11_virtual_base, module, NULL);
+	filter = p11_filter_subclass (&virt, NULL);
+	if (filter == NULL)
+		goto out;
+
+	for (i = 0; i < n_tokens; i++) {
+		P11KitUri *uri;
+		CK_TOKEN_INFO *token;
+
+		uri = p11_kit_uri_new ();
+		if (uri == NULL)
+			goto out;
+
+		if (p11_kit_uri_parse (tokens[i], P11_KIT_URI_FOR_TOKEN, uri) !=
+		    P11_KIT_URI_OK) {
+			p11_kit_uri_free (uri);
+			goto out;
+		}
+
+		token = p11_kit_uri_get_token_info (uri);
+		p11_filter_allow_token (filter, token);
+		p11_kit_uri_free (uri);
+	}
+
+	filtered = p11_virtual_wrap (filter, (p11_destroyer)p11_virtual_uninit);
+	if (filtered == NULL)
+		goto out;
+
+	ret = p11_kit_remote_serve_module (filtered, in_fd, out_fd);
+
+ out:
+	if (filtered != NULL)
+		p11_virtual_unwrap (filtered);
+	if (filter != NULL)
+		p11_filter_release (filter);
+
+	return ret;
+}
